from ._adapt import IS_NOSQL

if IS_NOSQL:
    from .nosql import *
else:
    from .sql import *
    from .indexes import *

from .base import *
from .caching import TestCache
from .contribs import *
from .is_url_validators import *
from .restapi import *
from .smart_query import *
from .tags import *
from .validation import *
from .validators import *
from .querybuilder import *
# from .scheduler import *



from ..adapters.sap import SAPDB
from . import dialects, sqltype_for
from .base import SQLDialect


@dialects.register_for(SAPDB)
class SAPDBDialect(SQLDialect):
    @sqltype_for("integer")
    def type_integer(self):
        return "INT"

    @sqltype_for("text")
    def type_text(self):
        return "LONG"

    @sqltype_for("bigint")
    def type_bigint(self):
        return "BIGINT"

    @sqltype_for("double")
    def type_double(self):
        return "DOUBLE PRECISION"

    @sqltype_for("decimal")
    def type_decimal(self):
        return "FIXED(%(precision)s,%(scale)s)"

    @sqltype_for("id")
    def type_id(self):
        return "INT PRIMARY KEY"

    @sqltype_for("big-id")
    def type_big_id(self):
        return "BIGINT PRIMARY KEY"

    @sqltype_for("reference")
    def type_reference(self):
        return (
            "INT, FOREIGN KEY (%(field_name)s) REFERENCES "
            + "%(foreign_key)s ON DELETE %(on_delete_action)s"
        )

    @sqltype_for("big-reference")
    def type_big_reference(self):
        return (
            "BIGINT, FOREIGN KEY (%(field_name)s) REFERENCES "
            + "%(foreign_key)s ON DELETE %(on_delete_action)s"
        )

    def sequence_name(self, tablename):
        return self.quote("%s_id_Seq" % tablename)

    def select(
        self,
        fields,
        tables,
        where=None,
        groupby=None,
        having=None,
        orderby=None,
        limitby=None,
        distinct=False,
        for_update=False,
    ):
        dst, whr, grp, order, limit, offset, upd = "", "", "", "", "", "", ""
        if distinct is True:
            dst = " DISTINCT"
        elif distinct:
            dst = " DISTINCT ON (%s)" % distinct
        if where:
            whr = " %s" % self.where(where)
        if groupby:
            grp = " GROUP BY %s" % groupby
            if having:
                grp += " HAVING %s" % having
        if orderby:
            order = " ORDER BY %s" % orderby
        if limitby:
            (lmin, lmax) = limitby
            if whr:
                whr2 = whr + " AND w_row > %i" % lmin
            else:
                whr2 = self.where("w_row > %i" % lmin)
            return (
                "SELECT%s %s FROM (SELECT w_tmp.*, ROWNO w_row FROM "
                + "(SELECT %s FROM %s%s%s%s) w_tmp WHERE ROWNO=%i) %s%s%s%s;"
                % (
                    dst,
                    fields,
                    fields,
                    tables,
                    whr,
                    grp,
                    order,
                    lmax,
                    tables,
                    whr2,
                    grp,
                    order,
                )
            )
        if for_update:
            upd = " FOR UPDATE"
        return "SELECT%s%s%s %s FROM %s%s%s%s%s;" % (
            dst,
            limit,
            offset,
            fields,
            tables,
            whr,
            grp,
            order,
            upd,
        )



import unittest

from pydal import DAL, Field, QueryBuilder


class TestQueryBuilder(unittest.TestCase):
    def test_query_builder(self):
        db = DAL("sqlite:memory")
        db.define_table("thing", Field("name"), Field("solid", "boolean"))
        builder = QueryBuilder(db.thing)
        query = builder.parse("name is null")
        self.assertEqual(str(query), '("thing"."name" IS NULL)')
        query = builder.parse("name is not null")
        self.assertEqual(str(query), '("thing"."name" IS NOT NULL)')
        query = builder.parse("solid is true")
        self.assertEqual(str(query), '("thing"."solid" = \'T\')')
        query = builder.parse("solid is false")
        self.assertEqual(str(query), '("thing"."solid" = \'F\')')
        query = builder.parse("name is Chair")
        self.assertEqual(str(query), '("thing"."name" = \'Chair\')')
        query = builder.parse("name is equal to Chair")
        self.assertEqual(str(query), '("thing"."name" = \'Chair\')')
        query = builder.parse("name == Chair")
        self.assertEqual(str(query), '("thing"."name" = \'Chair\')')
        query = builder.parse('name == "Chair"')
        self.assertEqual(str(query), '("thing"."name" = \'Chair\')')
        query = builder.parse('name == "Cha\\"ir"')
        self.assertEqual(str(query), '("thing"."name" = \'Cha\\"ir\')')
        query = builder.parse('name != "Chair"')
        self.assertEqual(str(query), '("thing"."name" <> \'Chair\')')
        query = builder.parse('name < "Chair"')
        self.assertEqual(str(query), '("thing"."name" < \'Chair\')')
        query = builder.parse('name > "Chair"')
        self.assertEqual(str(query), '("thing"."name" > \'Chair\')')
        query = builder.parse('name <= "Chair"')
        self.assertEqual(str(query), '("thing"."name" <= \'Chair\')')
        query = builder.parse('name >= "Chair"')
        self.assertEqual(str(query), '("thing"."name" >= \'Chair\')')
        query = builder.parse("name in Chair, Table")
        self.assertEqual(str(query), "(\"thing\".\"name\" IN ('Chair','Table'))")
        query = builder.parse("name belongs Chair, Table")
        self.assertEqual(str(query), "(\"thing\".\"name\" IN ('Chair','Table'))")
        query = builder.parse('name belongs "Chair", "Table"')
        self.assertEqual(str(query), '("thing"."name" IN (\'Chair", "Table\'))')
        query = builder.parse('name contains "Chair"')
        self.assertEqual(
            str(query), "(LOWER(\"thing\".\"name\") LIKE '%chair%' ESCAPE '\\')"
        )
        query = builder.parse('name startswith "Chair"')
        self.assertEqual(str(query), "(\"thing\".\"name\" LIKE 'Chair%' ESCAPE '\\')")
        query = builder.parse('name starts with "Chair"')
        self.assertEqual(str(query), "(\"thing\".\"name\" LIKE 'Chair%' ESCAPE '\\')")
        query = builder.parse("name lower == chair")
        self.assertEqual(str(query), '(LOWER("thing"."name") = \'chair\')')
        query = builder.parse("name lower is equal to chair")
        self.assertEqual(str(query), '(LOWER("thing"."name") = \'chair\')')
        query = builder.parse("name upper == CHAIR")
        self.assertEqual(str(query), '(UPPER("thing"."name") = \'CHAIR\')')
        query = builder.parse("not name == Chair")
        self.assertEqual(str(query), '(NOT ("thing"."name" = \'Chair\'))')

        query = builder.parse("not (name == Chair)")
        self.assertEqual(str(query), '(NOT ("thing"."name" = \'Chair\'))')

        query = builder.parse("name == Chair or name is Table")
        self.assertEqual(
            str(query), '(("thing"."name" = \'Chair\') OR ("thing"."name" = \'Table\'))'
        )

        query = builder.parse("name == Chair and not name is Table")
        self.assertEqual(
            str(query),
            '(("thing"."name" = \'Chair\') AND (NOT ("thing"."name" = \'Table\')))',
        )

        query = builder.parse("not ((name == Chair) and not (name == Table))")
        self.assertEqual(
            str(query),
            '(NOT (("thing"."name" = \'Chair\') AND (NOT ("thing"."name" = \'Table\'))))',
        )

    def test_translations(self):
        db = DAL("sqlite:memory")
        db.define_table("thing", Field("name"))
        field_aliases = {"id": "id", "nome": "name"}
        token_aliases = {"non è nullo": "is not null", "è uguale a": "=="}
        builder = QueryBuilder(
            db.thing, field_aliases=field_aliases, token_aliases=token_aliases
        )

        query = builder.parse("nome non è nullo")
        self.assertEqual(str(query), '("thing"."name" IS NOT NULL)')
        query = builder.parse("nome è uguale a Chair")
        self.assertEqual(str(query), '("thing"."name" = \'Chair\')')



from collections import defaultdict

from .._compat import iteritems, with_metaclass
from ..helpers._internals import Dispatcher

parsers = Dispatcher("parser")


class for_type(object):
    def __init__(self, field_type):
        self.field_type = field_type

    def __call__(self, f):
        self.f = f
        return self


class before_parse(object):
    def __init__(self, field_type):
        self.field_type = field_type

    def __call__(self, f):
        self.f = f
        return self


class MetaParser(type):
    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        if bases == (object,):
            return new_class
        #: collect declared attributes
        parsers = {}
        before = {}
        for key, value in list(attrs.items()):
            if isinstance(value, for_type):
                parsers[key] = value
            elif isinstance(value, before_parse):
                before[key] = value
        #: get super declared attributes
        declared_parsers = {}
        declared_before = {}
        for base in reversed(new_class.__mro__[1:]):
            if hasattr(base, "_declared_parsers_"):
                declared_parsers.update(base._declared_parsers_)
            if hasattr(base, "_declared_before_"):
                declared_before.update(base._declared_before_)
        #: set parsers
        declared_parsers.update(parsers)
        declared_before.update(before)
        new_class._declared_parsers_ = declared_parsers
        new_class._declared_before_ = declared_before
        return new_class


class ParserMethodWrapper(object):
    def __init__(self, parser, f, extra=None):
        self.parser = parser
        self.f = f
        if extra:
            self.extra = extra
            self.call = self._call_with_extras
        else:
            self.call = self._call

    def _call_with_extras(self, value, field_type):
        extras = self.extra(self.parser, field_type)
        return self.f(self.parser, value, **extras)

    def _call(self, value, field_type):
        return self.f(self.parser, value)

    def __call__(self, value, field_type):
        return self.call(value, field_type)


class Parser(with_metaclass(MetaParser)):
    def __init__(self, adapter):
        self.adapter = adapter
        self.dialect = adapter.dialect
        self._before_registry_ = {}
        for name, obj in iteritems(self._declared_before_):
            self._before_registry_[obj.field_type] = obj.f
        self.registered = defaultdict(lambda self=self: self._default)
        for name, obj in iteritems(self._declared_parsers_):
            if obj.field_type in self._before_registry_:
                self.registered[obj.field_type] = ParserMethodWrapper(
                    self, obj.f, self._before_registry_[obj.field_type]
                )
            else:
                self.registered[obj.field_type] = ParserMethodWrapper(self, obj.f)

    def _default(self, value, field_type):
        return value

    def parse(self, value, field_itype, field_type):
        return self.registered[field_itype](value, field_type)


from .base import BasicParser
from .google import FirestoreParser
from .mongo import MongoParser
from .oracle import OracleParser
from .postgre import PostgreParser
from .sqlite import SQLiteParser



__version__ = "20250526.4"

from .base import DAL
from .helpers.classes import SQLCustomType
from .helpers.methods import geoLine, geoPoint, geoPolygon
from .objects import Field
from .querybuilder import QueryBuilder



from ..adapters.google import Firestore
from . import dialects, sqltype_for
from .base import NoSQLDialect

try:
    from google.cloud import firestore
    from google.cloud.firestore_v1.base_query import (
        BaseCompositeFilter,
        FieldFilter,
        Or,
    )
except ImportError:
    pass


@dialects.register_for(Firestore)
class FirestoreDialect(NoSQLDialect):
    def _and(self, first, second, query_env={}):
        a = self.expand(first, query_env=query_env)
        b = self.expand(second, query_env=query_env)
        filters = (
            a.filters
            if isinstance(a, BaseCompositeFilter) and a.operator == "AND"
            else [a]
        )
        filters += (
            b.filters
            if isinstance(b, BaseCompositeFilter) and a.operator == "AND"
            else [b]
        )
        return BaseCompositeFilter("AND", filters)

    def _or(self, first, second, query_env={}):
        a = self.expand(first, query_env=query_env)
        b = self.expand(second, query_env=query_env)
        filters = (
            a.filters
            if isinstance(a, BaseCompositeFilter) and a.operator == "OR"
            else [a]
        )
        filters += (
            b.filters
            if isinstance(b, BaseCompositeFilter) and a.operator == "OR"
            else [b]
        )
        return BaseCompositeFilter("OR", filters)

    def eq(self, first, second=None, query_env={}):
        return FieldFilter(
            first.name, "==", self.expand(second, first.type, query_env=query_env)
        )

    def ne(self, first, second=None, query_env={}):
        return FieldFilter(
            first.name, "!=", self.expand(second, first.type, query_env=query_env)
        )

    def lt(self, first, second=None, query_env={}):
        return FieldFilter(
            first.name, "<", self.expand(second, first.type, query_env=query_env)
        )

    def lte(self, first, second=None, query_env={}):
        return FieldFilter(
            first.name, "<=", self.expand(second, first.type, query_env=query_env)
        )

    def gt(self, first, second=None, query_env={}):
        return FieldFilter(
            first.name, ">", self.expand(second, first.type, query_env=query_env)
        )

    def gte(self, first, second=None, query_env={}):
        return FieldFilter(
            first.name, ">=", self.expand(second, first.type, query_env=query_env)
        )

    def invert(self, first, query_env={}):
        raise NotImplementedError

    def comma(self, first, second, query_env={}):
        raise NotImplementedError

    def belongs(self, first, second, query_env={}):
        return FieldFilter(first.name, "in", second)

    def contains(self, first, second, case_sensitive=True, query_env={}):
        raise NotImplementedError

    def _not(self, val, query_env={}):
        op, f, s = val.op, val.first, val.second
        if op in [self._or, self._and]:
            not_op = self._and if op == self._or else self._or
            rv = not_op(self._not(f), self._not(s))
        elif op == self.eq:
            rv = FieldFilter(f.name, "!=", s)
        elif op == self.ne:
            rv = FieldFilter(f.name, "==", s)
        elif op == self.lt:
            rv = FieldFilter(f.name, ">=", s)
        elif op == self.lte:
            rv = FieldFilter(f.name, ">", s)
        elif op == self.gt:
            rv = FieldFilter(f.name, "<=", s)
        elif op == self.gte:
            rv = FieldFilter(f.name, "<", s)
        elif op == self.belongs:
            rv = FieldFilter(f.name, "not-in", s)
        else:
            # TODO the IN operator must be split into a sequence of
            # (field!=value) AND (field!=value) AND ...
            raise NotImplementedError
        return rv



from . import adapters, with_connection_or_raise
from .base import SQLAdapter


@adapters.register_for("informix")
class Informix(SQLAdapter):
    dbengine = "informix"
    drivers = ("informixdb",)

    def _initialize_(self):
        super(Informix, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        m = self.REGEX_URI.match(ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL")
        user = self.credential_decoder(m.group("user"))
        if not user:
            raise SyntaxError("User required")
        password = self.credential_decoder(m.group("password"))
        if not password:
            password = ""
        host = m.group("host")
        if not host:
            raise SyntaxError("Host name required")
        db = m.group("db")
        if not db:
            raise SyntaxError("Database name required")
        self.dsn = "%s@%s" % (db, host)
        self.driver_args.update(user=user, password=password)
        self.get_connection()

    def connector(self):
        return self.driver.connect(self.dsn, **self.driver_args)

    def _after_first_connection(self):
        self.dbms_version = int(self.connection.dbms_version.split(".")[0])

    @with_connection_or_raise
    def execute(self, *args, **kwargs):
        command = self.filter_sql_command(args[0])
        if command[-1:] == ";":
            command = command[:-1]
        handlers = self._build_handlers_for_execution()
        for handler in handlers:
            handler.before_execute(command)
        rv = self.cursor.execute(command, *args[1:], **kwargs)
        for handler in handlers:
            handler.after_execute(command)
        return rv

    def test_connection(self):
        self.execute("SELECT COUNT(*) FROM systables;")

    def lastrowid(self, table):
        return self.cursor.sqlerrd[1]


@adapters.register_for("informix-se")
class InformixSE(Informix):
    def rowslice(self, rows, minimum=0, maximum=None):
        if maximum is None:
            return rows[minimum:]
        return rows[minimum:maximum]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Unit tests for IS_URL()
"""

import unittest

from pydal.validators import (
    IS_GENERIC_URL,
    IS_HTTP_URL,
    IS_URL,
    unicode_to_ascii_authority,
)


class TestIsUrl(unittest.TestCase):
    def testModeHttp(self):
        # defaults to mode='http'

        x = IS_URL()
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(x("google.ca"), ("http://google.ca", None))
        self.assertEqual(x("google.ca:80"), ("http://google.ca:80", None))
        self.assertEqual(x("unreal.blargg"), ("unreal.blargg", "Enter a valid URL"))
        self.assertEqual(x("google..ca"), ("google..ca", "Enter a valid URL"))
        self.assertEqual(x("google.ca.."), ("google.ca..", "Enter a valid URL"))

        # explicit use of 'http' mode

        x = IS_URL(mode="http")
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(x("google.ca"), ("http://google.ca", None))
        self.assertEqual(x("google.ca:80"), ("http://google.ca:80", None))
        self.assertEqual(x("unreal.blargg"), ("unreal.blargg", "Enter a valid URL"))

        # prepends 'https' instead of 'http'

        x = IS_URL(mode="http", prepend_scheme="https")
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(x("google.ca"), ("https://google.ca", None))
        self.assertEqual(x("google.ca:80"), ("https://google.ca:80", None))
        self.assertEqual(x("unreal.blargg"), ("unreal.blargg", "Enter a valid URL"))

        # prepending disabled

        x = IS_URL(prepend_scheme=None)
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(x("google.ca"), ("google.ca", None))
        self.assertEqual(x("google.ca:80"), ("google.ca:80", None))
        self.assertEqual(x("unreal.blargg"), ("unreal.blargg", "Enter a valid URL"))

        # custom allowed_schemes

        x = IS_URL(mode="http", allowed_schemes=[None, "http"])
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(
            x("https://google.ca"), ("https://google.ca", "Enter a valid URL")
        )
        self.assertEqual(x("google.ca"), ("http://google.ca", None))
        self.assertEqual(x("google.ca:80"), ("http://google.ca:80", None))
        self.assertEqual(x("unreal.blargg"), ("unreal.blargg", "Enter a valid URL"))

        # custom allowed_schemes, excluding None

        x = IS_URL(allowed_schemes=["http"])
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(
            x("https://google.ca"), ("https://google.ca", "Enter a valid URL")
        )
        self.assertEqual(x("google.ca"), ("google.ca", "Enter a valid URL"))
        self.assertEqual(x("google.ca:80"), ("google.ca:80", "Enter a valid URL"))
        self.assertEqual(x("unreal.blargg"), ("unreal.blargg", "Enter a valid URL"))

        # custom allowed_schemes and prepend_scheme

        x = IS_URL(allowed_schemes=[None, "https"], prepend_scheme="https")
        self.assertEqual(
            x("http://google.ca"), ("http://google.ca", "Enter a valid URL")
        )
        self.assertEqual(x("https://google.ca"), ("https://google.ca", None))
        self.assertEqual(x("google.ca"), ("https://google.ca", None))
        self.assertEqual(x("google.ca:80"), ("https://google.ca:80", None))
        self.assertEqual(x("unreal.blargg"), ("unreal.blargg", "Enter a valid URL"))

        # Now any URL requiring prepending will fail, but prepending is still
        # enabled!

        x = IS_URL(allowed_schemes=["http"])
        self.assertEqual(x("google.ca"), ("google.ca", "Enter a valid URL"))

    def testModeGeneric(self):
        # 'generic' mode

        x = IS_URL(mode="generic")
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(x("google.ca"), ("google.ca", None))
        self.assertEqual(x("google.ca:80"), ("http://google.ca:80", None))
        self.assertEqual(x("blargg://unreal"), ("blargg://unreal", "Enter a valid URL"))

        # 'generic' mode with custom allowed_schemes that still includes
        # 'http' (the default for prepend_scheme)

        x = IS_URL(mode="generic", allowed_schemes=["http", "blargg"])
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(x("ftp://google.ca"), ("ftp://google.ca", "Enter a valid URL"))
        self.assertEqual(x("google.ca"), ("google.ca", "Enter a valid URL"))
        self.assertEqual(x("google.ca:80"), ("google.ca:80", "Enter a valid URL"))
        self.assertEqual(x("blargg://unreal"), ("blargg://unreal", None))

        # 'generic' mode with overriden prepend_scheme

        x = IS_URL(mode="generic", prepend_scheme="ftp")
        self.assertEqual(x("http://google.ca"), ("http://google.ca", None))
        self.assertEqual(x("ftp://google.ca"), ("ftp://google.ca", None))
        self.assertEqual(x("google.ca"), ("google.ca", None))
        self.assertEqual(x("google.ca:80"), ("ftp://google.ca:80", None))
        self.assertEqual(x("blargg://unreal"), ("blargg://unreal", "Enter a valid URL"))

        # 'generic' mode with overriden allowed_schemes and prepend_scheme

        x = IS_URL(
            mode="generic", allowed_schemes=[None, "ftp", "ftps"], prepend_scheme="ftp"
        )
        self.assertEqual(
            x("http://google.ca"), ("http://google.ca", "Enter a valid URL")
        )
        self.assertEqual(x("google.ca"), ("google.ca", None))
        self.assertEqual(x("ftp://google.ca"), ("ftp://google.ca", None))
        self.assertEqual(x("google.ca:80"), ("ftp://google.ca:80", None))
        self.assertEqual(x("blargg://unreal"), ("blargg://unreal", "Enter a valid URL"))

        # Now any URL requiring prepending will fail, but prepending is still
        # enabled!

        x = IS_URL(mode="generic", allowed_schemes=["http"])
        self.assertEqual(x("google.ca"), ("google.ca", "Enter a valid URL"))

    def testExceptionalUse(self):
        # mode must be in set ['http', 'generic']

        try:
            x = IS_URL(mode="ftp")
            x("http://www.google.ca")
        except Exception as e:
            if str(e) != "invalid mode 'ftp' in IS_URL":
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Accepted invalid mode: 'ftp'")

        # allowed_schemes in 'http' mode must be in set [None, 'http', 'https']

        try:
            x = IS_URL(allowed_schemes=[None, "ftp", "ftps"], prepend_scheme="ftp")
            x("http://www.benn.ca")  # we can only reasonably know about the
            # error at calling time
        except Exception as e:
            if str(e) != "allowed_scheme value 'ftp' is not in [None, 'http', 'https']":
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Accepted invalid allowed_schemes: [None, 'ftp', 'ftps']")

        # prepend_scheme's value must be in allowed_schemes (default for 'http'
        # mode is [None, 'http', 'https'])

        try:
            x = IS_URL(prepend_scheme="ftp")
            x("http://www.benn.ca")  # we can only reasonably know about the
            # error at calling time
        except Exception as e:
            if (
                str(e)
                != "prepend_scheme='ftp' is not in allowed_schemes=[None, 'http', 'https']"
            ):
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'ftp'")

        # custom allowed_schemes that excludes 'http', so prepend_scheme must be
        # specified!

        try:
            x = IS_URL(allowed_schemes=[None, "https"])
        except Exception as e:
            if (
                str(e)
                != "prepend_scheme='http' is not in allowed_schemes=[None, 'https']"
            ):
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'http'")

        # prepend_scheme must be in allowed_schemes

        try:
            x = IS_URL(allowed_schemes=[None, "http"], prepend_scheme="https")
        except Exception as e:
            if (
                str(e)
                != "prepend_scheme='https' is not in allowed_schemes=[None, 'http']"
            ):
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'https'")

        # prepend_scheme's value (default is 'http') must be in allowed_schemes

        try:
            x = IS_URL(mode="generic", allowed_schemes=[None, "ftp", "ftps"])
        except Exception as e:
            if (
                str(e)
                != "prepend_scheme='http' is not in allowed_schemes=[None, 'ftp', 'ftps']"
            ):
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'http'")

        # prepend_scheme's value must be in allowed_schemes, which by default
        # is all schemes that really exist

        try:
            x = IS_URL(mode="generic", prepend_scheme="blargg")
            x("http://www.google.ca")
            # we can only reasonably know about the error at calling time
        except Exception as e:
            if not str(e).startswith(
                "prepend_scheme='blargg' is not in allowed_schemes="
            ):
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'blargg'")

        # prepend_scheme's value must be in allowed_schemes

        try:
            x = IS_URL(
                mode="generic", allowed_schemes=[None, "http"], prepend_scheme="blargg"
            )
        except Exception as e:
            if (
                str(e)
                != "prepend_scheme='blargg' is not in allowed_schemes=[None, 'http']"
            ):
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Accepted invalid prepend_scheme: 'blargg'")

        # Not inluding None in the allowed_schemes essentially disabled
        # prepending, so even though
        # prepend_scheme has the invalid value 'http', we don't care!

        x = IS_URL(allowed_schemes=["https"], prepend_scheme="https")
        self.assertEqual(x("google.ca"), ("google.ca", "Enter a valid URL"))

        # Not inluding None in the allowed_schemes essentially disabled prepending, so even though
        # prepend_scheme has the invalid value 'http', we don't care!

        x = IS_URL(mode="generic", allowed_schemes=["https"], prepend_scheme="https")
        self.assertEqual(x("google.ca"), ("google.ca", "Enter a valid URL"))


# ##############################################################################


class TestIsGenericUrl(unittest.TestCase):
    x = IS_GENERIC_URL()

    def testInvalidUrls(self):
        urlsToCheckA = []
        for i in list(range(0, 32)) + [127]:
            # Control characters are disallowed in any part of a URL

            urlsToCheckA.append("http://www.benn" + chr(i) + ".ca")

        urlsToCheckB = [
            None,
            "",
            "http://www.no spaces allowed.com",
            "http://www.benn.ca/no spaces allowed/",
            "http://www.benn.ca/angle_<bracket/",
            "http://www.benn.ca/angle_>bracket/",
            "http://www.benn.ca/invalid%character",
            "http://www.benn.ca/illegal%%20use",
            "http://www.benn.ca/illegaluse%",
            "http://www.benn.ca/illegaluse%0",
            "http://www.benn.ca/illegaluse%x",
            "http://www.benn.ca/ill%egaluse%x",
            'http://www.benn.ca/double"quote/',
            "http://www.curly{brace.com",
            "http://www.benn.ca/curly}brace/",
            "http://www.benn.ca/or|symbol/",
            "http://www.benn.ca/back\\slash",
            "http://www.benn.ca/the^carat",
            "http://left[bracket.me",
            "http://www.benn.ca/right]bracket",
            "http://www.benn.ca/angle`quote",
            "-ttp://www.benn.ca",
            "+ttp://www.benn.ca",
            ".ttp://www.benn.ca",
            "9ttp://www.benn.ca",
            "ht;tp://www.benn.ca",
            "ht@tp://www.benn.ca",
            "ht&tp://www.benn.ca",
            "ht=tp://www.benn.ca",
            "ht$tp://www.benn.ca",
            "ht,tp://www.benn.ca",
            "ht:tp://www.benn.ca",
            "htp://invalid_scheme.com",
        ]

        failures = []

        for url in urlsToCheckA + urlsToCheckB:
            if self.x(url)[1] is None:
                failures.append("Incorrectly accepted: " + str(url))

        if len(failures) > 0:
            self.fail(failures)

    def testValidUrls(self):
        urlsToCheck = [
            "ftp://ftp.is.co.za/rfc/rfc1808.txt",
            "gopher://spinaltap.micro.umn.edu/00/Weather/California/Los%20Angeles",
            "http://www.math.uio.no/faq/compression-faq/part1.html",
            "mailto:mduerst@ifi.unizh.ch",
            "news:comp.infosystems.www.servers.unix",
            "telnet://melvyl.ucop.edu/",
            "hTTp://www.benn.ca",
            "%66%74%70://ftp.is.co.za/rfc/rfc1808.txt",
            "%46%74%70://ftp.is.co.za/rfc/rfc1808.txt",
            "/faq/compression-faq/part1.html",
            "google.com",
            "www.google.com:8080",
            "128.127.123.250:8080",
            "blargg:ping",
            "http://www.benn.ca",
            "http://benn.ca",
            "http://amazon.com/books/",
            "https://amazon.com/movies",
            "rtsp://idontknowthisprotocol",
            "HTTP://allcaps.com",
            "http://localhost",
            "http://localhost#fragment",
            "http://localhost/hello",
            "http://localhost/hello?query=True",
            "http://localhost/hello/",
            "http://localhost:8080",
            "http://localhost:8080/",
            "http://localhost:8080/hello",
            "http://localhost:8080/hello/",
            "file:///C:/Documents%20and%20Settings/Jonathan/Desktop/view.py",
        ]

        failures = []

        for url in urlsToCheck:
            if self.x(url)[1] is not None:
                failures.append("Incorrectly rejected: " + str(url))

        if len(failures) > 0:
            self.fail(failures)

    def testPrepending(self):
        # Does not prepend scheme for abbreviated domains
        self.assertEqual(self.x("google.ca"), ("google.ca", None))

        # Does not prepend scheme for abbreviated domains
        self.assertEqual(self.x("google.ca:8080"), ("google.ca:8080", None))

        # Does not prepend when scheme already exists
        self.assertEqual(self.x("https://google.ca"), ("https://google.ca", None))

        # Does not prepend if None type is not specified in allowed_scheme,
        # because a scheme is required

        y = IS_GENERIC_URL(allowed_schemes=["http", "blargg"], prepend_scheme="http")
        self.assertEqual(y("google.ca"), ("google.ca", "Enter a valid URL"))


# ##############################################################################


class TestIsHttpUrl(unittest.TestCase):
    x = IS_HTTP_URL()

    def testInvalidUrls(self):
        urlsToCheck = [
            None,
            "",
            "http://invalid" + chr(2) + ".com",
            "htp://invalid_scheme.com",
            "blargg://invalid_scheme.com",
            "http://-123.com",
            "http://abcd-.ca",
            "http://-abc123-.me",
            "http://www.dom&ain.com/",
            "http://www.dom=ain.com/",
            "http://www.benn.ca&",
            "http://%62%65%6E%6E%2E%63%61/path",
            "http://.domain.com",
            "http://.domain.com./path",
            "http://domain..com",
            "http://domain...at..com",
            "http://domain.com..",
            "http://domain.com../path",
            "http://domain.3m",
            "http://domain.-3m",
            "http://domain.3m-",
            "http://domain.-3m-",
            "http://domain.co&m",
            "http://domain.m3456",
            "http://domain.m-3/path#fragment",
            "http://domain.m---k/path?query=value",
            "http://23.32..",
            "http://23..32.56.0",
            "http://38997.222.999",
            "http://23.32.56.99.",
            "http://.23.32.56.99",
            "http://.23.32.56.99.",
            "http://w127.123.0.256:8080",
            "http://23.32.56.99:abcd",
            "http://23.32.56.99:23cd",
            "http://google.com:cd22",
            "http://23.32:1300.56.99",
            "http://www.yahoo:1600.com",
            "path/segment/without/starting/slash",
            "http://www.math.uio.no;param=3",
            "://ABC.com:/%7esmith/home.html",
        ]

        failures = []

        for url in urlsToCheck:
            if self.x(url)[1] is None:
                failures.append("Incorrectly accepted: " + str(url))

        if len(failures) > 0:
            self.fail(failures)

    def testValidUrls(self):
        urlsToCheck = [
            "http://abc.com:80/~smith/home.html",
            "http://ABC.com/%7Esmith/home.html",
            "http://ABC.com:/%7esmith/home.html",
            "http://www.math.uio.no/faq/compression-faq/part1.html",
            "//google.ca/faq/compression-faq/part1.html",
            "//google.ca/faq;param=3",
            "//google.ca/faq/index.html?query=5",
            "//google.ca/faq/index.html;param=value?query=5",
            "/faq/compression-faq/part1.html",
            "/faq;param=3",
            "/faq/index.html?query=5",
            "/faq/index.html;param=value?query=5",
            "google.com",
            "benn.ca/init/default",
            "benn.ca/init;param=value/default?query=value",
            "http://host-name---with-dashes.me",
            "http://www.host-name---with-dashes.me",
            "http://a.com",
            "http://a.3.com",
            "http://a.bl-ck.com",
            "http://bl-e.b.com",
            "http://host123with456numbers.ca",
            "http://1234567890.com.",
            "http://1234567890.com./path",
            "http://google.com./path",
            "http://domain.xn--d1acj3b",
            "http://127.123.0.256",
            "http://127.123.0.256/document/drawer",
            "127.123.0.256/document/",
            "156.212.123.100",
            "http://www.google.com:180200",
            "http://www.google.com:8080/path",
            "http://www.google.com:8080",
            "//www.google.com:8080",
            "www.google.com:8080",
            "http://127.123.0.256:8080/path",
            "//127.123.0.256:8080",
            "127.123.0.256:8080",
            "http://example.me??query=value?",
            "http://a.com",
            "http://3.com",
            "http://www.benn.ca",
            "http://benn.ca",
            "http://amazon.com/books/",
            "https://amazon.com/movies",
            "hTTp://allcaps.com",
            "http://localhost",
            "HTTPS://localhost.",
            "http://localhost#fragment",
            "http://localhost/hello;param=value",
            "http://localhost/hello;param=value/hi;param2=value2;param3=value3",
            "http://localhost/hello?query=True",
            "http://www.benn.ca/hello;param=value/hi;param2=value2;param3=value3/index.html?query=3",
            "http://localhost/hello/?query=1500&five=6",
            "http://localhost:8080",
            "http://localhost:8080/",
            "http://localhost:8080/hello",
            "http://localhost:8080/hello%20world/",
            "http://www.a.3.be-nn.5.ca",
            "http://www.amazon.COM",
        ]

        failures = []

        for url in urlsToCheck:
            if self.x(url)[1] is not None:
                failures.append("Incorrectly rejected: " + str(url))

        if len(failures) > 0:
            self.fail(failures)

    def testPrepending(self):
        # prepends scheme for abbreviated domains
        self.assertEqual(self.x("google.ca"), ("http://google.ca", None))

        # prepends scheme for abbreviated domains
        self.assertEqual(self.x("google.ca:8080"), ("http://google.ca:8080", None))

        # does not prepend when scheme already exists
        self.assertEqual(self.x("https://google.ca"), ("https://google.ca", None))

        y = IS_HTTP_URL(prepend_scheme="https", allowed_schemes=[None, "https"])
        self.assertEqual(
            y("google.ca"), ("https://google.ca", None)
        )  # prepends https if asked

        z = IS_HTTP_URL(prepend_scheme=None)
        self.assertEqual(
            z("google.ca:8080"), ("google.ca:8080", None)
        )  # prepending disabled

        try:
            IS_HTTP_URL(prepend_scheme="mailto")
        except Exception as e:
            if (
                str(e)
                != "prepend_scheme='mailto' is not in allowed_schemes=[None, 'http', 'https']"
            ):
                self.fail("Wrong exception: " + str(e))
        else:
            self.fail("Got invalid prepend_scheme: 'mailto'")

        # Does not prepend if None type is not specified in allowed_scheme, because a scheme is required

        a = IS_HTTP_URL(allowed_schemes=["http"])
        self.assertEqual(a("google.ca"), ("google.ca", "Enter a valid URL"))
        self.assertEqual(a("google.ca:80"), ("google.ca:80", "Enter a valid URL"))


class TestUnicode(unittest.TestCase):
    x = IS_URL()
    y = IS_URL(allowed_schemes=["https"], prepend_scheme="https")
    # excludes the option for abbreviated URLs with no scheme
    z = IS_URL(prepend_scheme=None)
    # disables prepending the scheme in the return value

    def testUnicodeToAsciiUrl(self):
        self.assertEqual(
            unicode_to_ascii_authority("www.Alliancefran\xe7aise.nu"),
            "www.xn--alliancefranaise-npb.nu",
        )
        self.assertEqual(unicode_to_ascii_authority("www.benn.ca"), "www.benn.ca")
        self.assertRaises(
            UnicodeError, unicode_to_ascii_authority, "\u4e2d" * 1000
        )  # label is too long

    def testValidUrls(self):
        self.assertEqual(
            self.x("www.Alliancefrancaise.nu"),
            ("http://www.Alliancefrancaise.nu", None),
        )
        self.assertEqual(
            self.x("www.Alliancefran\xe7aise.nu"),
            ("http://www.xn--alliancefranaise-npb.nu", None),
        )
        self.assertEqual(
            self.x("www.Alliancefran\xe7aise.nu:8080"),
            ("http://www.xn--alliancefranaise-npb.nu:8080", None),
        )
        self.assertEqual(
            self.x("http://www.Alliancefran\xe7aise.nu"),
            ("http://www.xn--alliancefranaise-npb.nu", None),
        )
        self.assertEqual(
            self.x("http://www.Alliancefran\xe7aise.nu/parnaise/blue"),
            ("http://www.xn--alliancefranaise-npb.nu/parnaise/blue", None),
        )
        self.assertEqual(
            self.x("http://www.Alliancefran\xe7aise.nu/parnaise/blue#fragment"),
            ("http://www.xn--alliancefranaise-npb.nu/parnaise/blue#fragment", None),
        )
        self.assertEqual(
            self.x(
                "http://www.Alliancefran\xe7aise.nu/parnaise/blue?query=value#fragment"
            ),
            (
                "http://www.xn--alliancefranaise-npb.nu/parnaise/blue?query=value#fragment",
                None,
            ),
        )
        self.assertEqual(
            self.x(
                "http://www.Alliancefran\xe7aise.nu:8080/parnaise/blue?query=value#fragment"
            ),
            (
                "http://www.xn--alliancefranaise-npb.nu:8080/parnaise/blue?query=value#fragment",
                None,
            ),
        )
        self.assertEqual(
            self.x("www.Alliancefran\xe7aise.nu/parnaise/blue?query=value#fragment"),
            (
                "http://www.xn--alliancefranaise-npb.nu/parnaise/blue?query=value#fragment",
                None,
            ),
        )
        self.assertEqual(
            self.x("http://\u4e2d\u4fd4.com"), ("http://xn--fiq13b.com", None)
        )
        self.assertEqual(
            self.x("http://\u4e2d\u4fd4.com/\u4e86"),
            ("http://xn--fiq13b.com/%4e%86", None),
        )
        self.assertEqual(
            self.x("http://\u4e2d\u4fd4.com/\u4e86?query=\u4e86"),
            ("http://xn--fiq13b.com/%4e%86?query=%4e%86", None),
        )
        self.assertEqual(
            self.x("http://\u4e2d\u4fd4.com/\u4e86?query=\u4e86#fragment"),
            ("http://xn--fiq13b.com/%4e%86?query=%4e%86#fragment", None),
        )
        self.assertEqual(
            self.x("http://\u4e2d\u4fd4.com?query=\u4e86#fragment"),
            ("http://xn--fiq13b.com?query=%4e%86#fragment", None),
        )
        self.assertEqual(
            self.x("http://B\xfccher.ch"), ("http://xn--bcher-kva.ch", None)
        )
        self.assertEqual(
            self.x("http://\xe4\xf6\xfc\xdf.com"), ("http://xn--ss-uia6e4a.com", None)
        )
        self.assertEqual(
            self.x("http://visegr\xe1d.com"), ("http://xn--visegrd-mwa.com", None)
        )
        self.assertEqual(
            self.x("http://h\xe1zipatika.com"), ("http://xn--hzipatika-01a.com", None)
        )
        self.assertEqual(
            self.x("http://www.\xe7ukurova.com"),
            ("http://www.xn--ukurova-txa.com", None),
        )
        self.assertEqual(
            self.x("http://nixier\xf6hre.nixieclock-tube.com"),
            ("http://xn--nixierhre-57a.nixieclock-tube.com", None),
        )
        self.assertEqual(self.x("google.ca."), ("http://google.ca.", None))

        self.assertEqual(self.y("https://google.ca"), ("https://google.ca", None))
        self.assertEqual(
            self.y("https://\u4e2d\u4fd4.com"), ("https://xn--fiq13b.com", None)
        )

        self.assertEqual(self.z("google.ca"), ("google.ca", None))

    def testInvalidUrls(self):
        self.assertEqual(self.x("://ABC.com"), ("://ABC.com", "Enter a valid URL"))
        self.assertEqual(
            self.x("http://\u4e2d\u4fd4.dne"),
            ("http://\u4e2d\u4fd4.dne", "Enter a valid URL"),
        )
        self.assertEqual(
            self.x("https://google.dne"), ("https://google.dne", "Enter a valid URL")
        )
        self.assertEqual(
            self.x("https://google..ca"), ("https://google..ca", "Enter a valid URL")
        )
        self.assertEqual(self.x("google..ca"), ("google..ca", "Enter a valid URL"))
        self.assertEqual(
            self.x("http://" + "\u4e2d" * 1000 + ".com"),
            ("http://" + "\u4e2d" * 1000 + ".com", "Enter a valid URL"),
        )

        self.assertEqual(
            self.x("http://google.com#fragment_\u4e86"),
            ("http://google.com#fragment_\u4e86", "Enter a valid URL"),
        )
        self.assertEqual(
            self.x("http\u4e86://google.com"),
            ("http\u4e86://google.com", "Enter a valid URL"),
        )
        self.assertEqual(
            self.x("http\u4e86://google.com#fragment_\u4e86"),
            ("http\u4e86://google.com#fragment_\u4e86", "Enter a valid URL"),
        )

        self.assertEqual(
            self.y("http://\u4e2d\u4fd4.com/\u4e86"),
            ("http://\u4e2d\u4fd4.com/\u4e86", "Enter a valid URL"),
        )
        # self.assertEqual(self.y(u'google.ca'), (u'google.ca', 'Enter a valid URL'))

        self.assertEqual(
            self.z("invalid.domain..com"),
            ("invalid.domain..com", "Enter a valid URL"),
        )
        self.assertEqual(
            self.z("invalid.\u4e2d\u4fd4.blargg"),
            ("invalid.\u4e2d\u4fd4.blargg", "Enter a valid URL"),
        )


# ##############################################################################


class TestSimple(unittest.TestCase):
    def test_IS_URL(self):
        rtn = IS_URL()("abc.com")
        self.assertEqual(rtn, ("http://abc.com", None))
        rtn = IS_URL(mode="generic")("abc.com")
        self.assertEqual(rtn, ("abc.com", None))
        rtn = IS_URL(allowed_schemes=["https"], prepend_scheme="https")(
            "https://abc.com"
        )
        self.assertEqual(rtn, ("https://abc.com", None))
        rtn = IS_URL(prepend_scheme="https")("abc.com")
        self.assertEqual(rtn, ("https://abc.com", None))
        rtn = IS_URL(
            mode="generic", allowed_schemes=["ftps", "https"], prepend_scheme="https"
        )("https://abc.com")
        self.assertEqual(rtn, ("https://abc.com", None))
        rtn = IS_URL(
            mode="generic",
            allowed_schemes=["ftps", "https", None],
            prepend_scheme="https",
        )("abc.com")
        self.assertEqual(rtn, ("abc.com", None))
        # regression test for issue 773
        rtn = IS_URL()("domain.ninja")
        self.assertEqual(rtn, ("http://domain.ninja", None))
        # addition of allowed_tlds
        rtn = IS_URL(allowed_tlds=["com", "net", "org"])("domain.ninja")
        self.assertEqual(rtn, ("domain.ninja", "Enter a valid URL"))
        # mode = 'generic' doesn't consider allowed_tlds
        rtn = IS_URL(mode="generic", allowed_tlds=["com", "net", "org"])("domain.ninja")
        self.assertEqual(rtn, ("domain.ninja", None))



# pylint: disable=broad-exception-caught,line-too-long,singleton-comparison,too-many-arguments,unnecessary-lambda
"""
A scheduler can can run any Python task in parallel using processes.
This file is part of pydal.
Created by Massimo Di Pierro<massimo.dipierro@gmail.com>
License: 3-clause BSD
"""

import datetime
import logging
import math
import os
import signal
import socket
import sys
import threading
import time
import traceback

from pydal import DAL, Field
from pydal.validators import IS_IN_SET

from ..utils import utcnow


def now():
    """Returns the current datetime in UTC"""
    return utcnow()


def delta(t_secs):
    """Returns a timedelta"""
    return datetime.timedelta(seconds=t_secs)


def make_daemon(func, filename, cwd="."):
    """Creates a daemon process running func in cwd and stdout->filename"""
    if os.fork():
        return
    # decouple from parent environment
    os.chdir(cwd)
    # os.setsid()
    os.umask(0)
    # do second fork
    if os.fork():
        sys.exit(0)
    # redirect standard file descriptors
    sys.__stdout__.flush()
    sys.__stderr__.flush()
    with open(os.devnull, "rb") as stream_in:
        os.dup2(stream_in.fileno(), sys.__stdin__.fileno())
        with open(filename, "wb") as stream_out:
            os.dup2(stream_out.fileno(), sys.__stdout__.fileno())
            os.dup2(stream_out.fileno(), sys.__stderr__.fileno())
            try:
                func()
            finally:
                stream_out.flush()
    sys.exit(0)


def pid_exists(pid):
    """Check For the existence of a unix pid."""
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def get_logger(name, log_format="[%(asctime)s]%(levelname)s %(message)s"):
    """Get a default logger"""
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.__stdout__)
    handler.setFormatter(logging.Formatter(log_format))
    logger.addHandler(handler)
    return logger


class Scheduler:  # pylint: disable=too-many-instance-attributes
    """Makes a scheduler"""

    __slots__ = [
        "tasks",
        "db",
        "max_concurrent_runs",
        "folder",
        "logger",
        "sleep_time",
        "worker",
        "_looping",
        "_thread",
    ]

    statuses = [
        "queued",  # the task has been queued and waiting to run
        "assigned",  # the task was assigned to a worker process
        "running",  # the task it running
        "timeout",  # the task took to long and was killed
        "completed",  # the task successfully completed
        "failed",  # the task failed
        "dead",  # the task no longer appear to be running
        "unknown",  # the task was queued but its name not in self.tasks
    ]

    def __init__(
        self,
        db,
        max_concurrent_runs=2,
        folder="/tmp/scheduler",
        sleep_time=10,
        logger=None,
    ):
        signal.signal(signal.SIGCHLD, signal.SIG_IGN)
        self.db = db
        self.max_concurrent_runs = max_concurrent_runs
        self.folder = folder
        self.sleep_time = sleep_time
        self.worker = socket.gethostbyname(socket.gethostname())
        self.logger = logger or get_logger("scheduler")
        self.tasks = {}
        self._looping = False
        self._thread = None
        # make the output folder if does not exist
        os.makedirs(self.folder, exist_ok=True)
        # create a task_run table if not already provided
        if "task_run" not in db:
            db.define_table(
                "task_run",
                # given
                Field(
                    "name", requires=IS_IN_SET(lambda: self.tasks.keys()), writable=True
                ),
                Field("description", "text", writable=True),
                Field("inputs", "json", default={}, writable=True),
                Field("timeout", "integer", default=None, writable=True),
                Field("priority", "integer", default=0, writable=True),
                Field("period", "integer", default=None, writable=True),
                Field("scheduled_for", "datetime", default=now, writable=True),
                # given but reassigned
                Field(
                    "status",
                    default="queued",
                    writable=False,
                    requires=IS_IN_SET(Scheduler.statuses),
                ),
                # computed
                Field("queued_on", "datetime", default=now, writable=False),
                Field("worker", writable=False),
                Field("pid", "integer", writable=False),
                Field("started_on", "datetime", writable=False),
                Field("completed_on", "datetime", writable=False),
                Field("log", "text", writable=False),
                Field("output", "json", writable=False),
            )
            db.commit()

    def start(self):
        """Starts a thread running the main loop of the scheduler"""
        assert not self._looping and self._thread is None
        self._looping = True
        self._thread = threading.Thread(target=self.loop)
        self._thread.start()

    def stop(self):
        """Stops the main loop and joins its thread"""
        assert self._thread and self._thread is not None
        self._looping = False
        self._thread.join()
        self._thread = None

    def has_work(self):
        """Checks whether the scheduler has any work queued, assigned, or running"""
        db = self.db
        return (
            db(db.task_run.status.belongs(("queued", "assigned", "running"))).count()
            > 0
        )

    def loop(self):
        """Runs the main loop of the scheduler"""
        db = self.db
        self.logger.info("worker %s/%s start", self.worker, id(self))
        while self._looping:
            self.logger.info("loop start (tasks in db %i)", db(db.task_run).count())
            if not self.step():
                time.sleep(self.sleep_time)
            self.logger.info("end loop")
        self.logger.info("worker %s/%s stop", self.worker, id(self))

    def step(self):
        """Runs one step of the scheduler"""
        db = self.db
        # find corrupted runs and re-enqueue them
        wruns = db(db.task_run.worker == self.worker)
        wruns(db.task_run.status == "assigned").update(status="queued")
        db.commit()
        # check on task timeout
        runs = wruns(db.task_run.status == "running").select()
        num_running = 0
        t_end = now()
        for run in runs:
            # check for processes that died
            if not pid_exists(run.pid):
                log = self.retrieve_log(run) + self._end_tag(
                    status="dead", completed_on=t_end
                )
                run.update_record(status="dead", log=log, completed_on=t_end)
                db.commit()
                self.logger.info("run died #%i %s", run.id, run.name)
            # check for processes that timedout
            if run.timeout and run.started_on + delta(run.timeout) < now():
                os.kill(run.pid, signal.SIGKILL)
                log = self.retrieve_log(run) + self._end_tag(
                    status="timeout", completed_on=t_end
                )
                run.update_record(status="timeout", log=log, completed_on=t_end)
                db.commit()
                self.logger.info("run timeout #%i %s", run.id, run.name)
            else:
                num_running += 1

        # if too main runs wait before retrying
        if num_running >= self.max_concurrent_runs:
            self.logger.info("too many running tasks")
            return False

        # try find a new run else wait before retrying
        run = self.next_run()
        if not run:
            self.logger.info("no new tasks")
            return False

        # if the next is unknown, skip it
        self.logger.info("new run: %s", run.name)
        if run.name not in self.tasks:
            run.update_record(status="unknown")
            db.commit()
            return True

        # make a child and assign it the run
        filename = self.get_output_filename(run)
        try:
            make_daemon(lambda run=run: self.safe_exec_child(run), filename)
        except OSError as err:
            self.logger.error("Fork error: %s", err)
        except Exception:
            self.logger.error(traceback.format_exc())
        return True

    def next_run(self):
        """Returns the next run ready to be executed"""
        db = self.db
        # find the next task to execute
        nruns = db(db.task_run.worker == None)(db.task_run.scheduled_for <= now())
        orderby = db.task_run.priority | db.task_run.id
        run = nruns.select(orderby=orderby, limitby=(0, 1)).first()
        self.logger.info("run found #%s %s", run and run.id, run and run.name)
        if run:
            # try assign the run to self if no other process stole it
            try:
                if (
                    nruns(db.task_run.id == run.id).update(
                        worker=self.worker, status="assigned"
                    )
                    > 0
                ):
                    db.commit()
                    self.logger.info("run assigned #%i %s", run.id, run.name)
                    return run
            except Exception:
                # some other process stole it, rollback and loop to try again
                self.logger.warning(traceback.format_exc())
                db.rollback()
        # run not fund
        return None

    def safe_exec_child(self, run):
        """Run exec_child in a try check to reconnect/commit/rollback all databases"""
        try:
            self.db.get_connection_from_pool_or_new()
            self.exec_child(run)
            self.db.recycle_connection_in_pool_or_close("commit")
        except Exception:
            self.db.recycle_connection_in_pool_or_close("rollback")

    def exec_child(self, run):
        """Executes the run, to be called by the child process, any output here goes into log file"""
        db = self.db
        pid = os.getpid()
        started_on = now()
        # print the header in the output log file
        print(self._begin_tag(run, started_on))
        # update the record
        run.update_record(status="running", started_on=started_on, pid=pid)
        db.commit()
        self.logger.info("run started #%i %s", run.id, run.name)
        # run the task safely
        try:
            output = self.tasks[run.name](**run.inputs)
            status, tb = "completed", None
        except Exception:
            status, output, tb = "failed", None, traceback.format_exc()
        # record the completion
        completed_on = now()
        # retrieve the log and update the current task
        log = self.retrieve_log(run) + self._end_tag(status, completed_on, tb)
        run.update_record(
            status=status, output=output, log=log, completed_on=completed_on
        )
        db.commit()
        self.logger.info("run %s #%i %s", status, run.id, run.name)
        # if periodic task, reschedule it (mind errors below will not be logged)
        if run.period:
            wait_time = (run.completed_on - run.scheduled_for).total_seconds()
            dt = run.period * math.ceil(wait_time / run.period)
            scheduled_for = run.scheduled_for + delta(dt)
            self.enqueue_run(
                run.name,
                run.description,
                run.inputs,
                run.timeout,
                run.priority,
                run.period,
                scheduled_for,
            )

    def _begin_tag(self, run, started_on):
        """Generate the begin <run> tag for the output file"""
        return f'<run id="{run.id}" worker="{run.worker}" pid="{run.pid}" started_on="{started_on}">'

    def _end_tag(self, status, completed_on, tb=None):
        """Generate the end </run> tag for the output file"""
        msg = f"\n<traceback>\n{tb.strip()}\n</traceback>" if tb else ""
        return f'{msg}\n</run status="{status}" completed_on="{completed_on}">'

    def retrieve_log(self, run):
        """Retrieve the log for the run"""
        try:
            filename = self.get_output_filename(run)
            with open(filename, "rb") as stream:
                log = stream.read().decode("utf8", errors="ignore")
            os.unlink(filename)
        except Exception as err:
            log = f'<missing reason="{err}"/>'
        return log.strip()

    def get_output_filename(self, run):
        """Generate the fullname for the output log file"""
        return os.path.join(self.folder, f"{run.id}.txt")

    def register_task(self, task_name, function):
        """Register a new task, given a name and a function"""
        self.tasks[task_name] = function

    def enqueue_run(
        self,
        name,
        description="",
        inputs={},
        timeout=3600,
        priority=0,
        period=None,
        scheduled_for=None,
    ):
        """Stores a new db.task_run"""
        db = self.db
        assert name in set(dict(db.task_run.name.requires.options()))
        assert inputs is None or isinstance(inputs, dict)
        t = now()
        run_id = db.task_run.insert(
            name=name,
            description=description,
            inputs=inputs or {},
            timeout=timeout,
            priority=priority,
            period=period,
            scheduled_for=scheduled_for or t,
            queued_on=t,
            status="queued",
        )
        db.commit()
        self.logger.info("run enqueued %s", run_id)


def run_example():
    """Example of usage"""

    def mytask():
        print(f"hello world {datetime.datetime.now()}")
        return "Done!"

    db = DAL("sqlite://storage.sqlite")
    scheduler = Scheduler(db)
    db(db.task_run).delete()
    # a task that says hello
    scheduler.register_task("hello", mytask)
    scheduler.register_task("periodic", mytask)
    # a task that fails if x is 0
    scheduler.register_task("fail", lambda x: 1 / x)
    # task that takes 10 seconds
    scheduler.register_task("long", lambda: time.sleep(10))
    # enqueue a run for the fail task with input x=0 (for indeed fails)
    scheduler.enqueue_run(name="fail", inputs={"x": 0})
    # a run that takes 10s but with timeout of 1s so timeouts
    scheduler.enqueue_run(name="long", timeout=1)
    # a run scheduled to start in 20s
    scheduler.enqueue_run(name="hello", scheduled_for=now() + delta(20))
    # a regular run to start asap
    scheduler.enqueue_run(name="hello")
    # a run that should rerun every 10s
    scheduler.enqueue_run(name="periodic", period=10)
    db.commit()
    scheduler.start()
    while scheduler.has_work():
        # do not all more than 3 periodic tasks
        if db(db.task_run.name == "periodic").count() > 3:
            db(db.task_run.name == "periodic")(db.task_run.status == "queued").delete()
            db.commit()
        time.sleep(1)
    scheduler.stop()


if __name__ == "__main__":
    run_example()



import base64
import datetime
from base64 import b64encode

from .._compat import to_bytes, to_native
from ..adapters.oracle import Oracle
from . import representers
from .base import JSONRepresenter, SQLRepresenter


@representers.register_for(Oracle)
class OracleRepresenter(SQLRepresenter, JSONRepresenter):
    def exceptions(self, obj, field_type):
        if field_type == "blob":
            # ensure bytes-like before conversion
            if type(obj) != bytes:
                obj = to_bytes(obj)
            obj = to_native(b64encode(obj))
            # return "utl_raw.cast_to_raw('%s')" % obj
            return ":CLOB('%s')" % obj
        elif field_type == "date":
            if isinstance(obj, (datetime.date, datetime.datetime)):
                obj = obj.isoformat()[:10]
            else:
                obj = str(obj)
            return "to_date('%s','yyyy-mm-dd')" % obj
        elif field_type == "datetime":
            if isinstance(obj, datetime.datetime):
                obj = obj.isoformat()[:19].replace("T", " ")
            elif isinstance(obj, datetime.date):
                obj = obj.isoformat()[:10] + " 00:00:00"
            else:
                obj = str(obj)
            return "to_date('%s','yyyy-mm-dd hh24:mi:ss')" % obj
        return None



from ..adapters.sqlite import Spatialite, SQLite
from . import before_type, for_type, representers
from .base import JSONRepresenter, SQLRepresenter


@representers.register_for(SQLite)
class SQLiteRepresenter(SQLRepresenter, JSONRepresenter):
    pass


@representers.register_for(Spatialite)
class SpatialiteRepresenter(SQLRepresenter):
    @before_type("geometry")
    def geometry_extras(self, field_type):
        srid = 4326
        geotype, params = field_type[:-1].split("(")
        params = params.split(",")
        if len(params) >= 2:
            schema, srid = params[:2]
        return {"srid": srid}

    @for_type("geometry", adapt=False)
    def _geometry(self, value, srid):
        return "ST_GeomFromText('%s',%s)" % (value, srid)



import os
import threading
import time

from pydal._compat import to_bytes
from pydal.contrib.portalocker import (
    LOCK_EX,
    LockedFile,
    lock,
    read_locked,
    unlock,
    write_locked,
)

from ._adapt import IS_GAE
from ._compat import unittest


def tearDownModule():
    if os.path.isfile("test.txt"):
        os.unlink("test.txt")


class testPortalocker(unittest.TestCase):
    def test_LockedFile(self):
        f = LockedFile("test.txt", mode="wb")
        f.write(to_bytes("test ok"))
        f.close()
        f = LockedFile("test.txt", mode="rb")
        self.assertEqual(f.read(), to_bytes("test ok"))
        f.close()

    @unittest.skipIf(IS_GAE, "GAE has no locks")
    def test_openmultiple(self):
        t0 = time.time()

        def worker1():
            start = time.time()
            f1 = LockedFile("test.txt", mode="ab")
            time.sleep(2)
            f1.write(to_bytes("%s\t%s\n" % (start, time.time())))
            f1.close()

        f = LockedFile("test.txt", mode="wb")
        f.write(to_bytes(""))
        f.close()
        th = []
        for x in range(10):
            t1 = threading.Thread(target=worker1)
            th.append(t1)
            t1.start()
        for t in th:
            t.join()
        with open("test.txt") as g:
            content = g.read()

        results = [line.strip().split("\t") for line in content.split("\n") if line]
        # all started at more or less the same time
        starts = [1 for line in results if float(line[0]) - t0 < 1]
        ends = [line[1] for line in results]
        self.assertEqual(sum(starts), len(starts))
        # end - start is at least 2
        for line in results:
            self.assertTrue(float(line[1]) - float(line[0]) >= 2)
        # ends are not the same
        self.assertTrue(len(ends) == len(ends))

    @unittest.skipIf(IS_GAE, "GAE has no locks")
    def test_lock_unlock(self):
        def worker1(fh):
            time.sleep(2)
            unlock(fh)

        def worker2(fh):
            time.sleep(2)
            fh.close()

        f = open("test.txt", mode="wb")
        lock(f, LOCK_EX)
        f.write(to_bytes("test ok"))
        t1 = threading.Thread(target=worker1, args=(f,))
        t1.start()
        start = int(time.time())
        content = read_locked("test.txt")
        end = int(time.time())
        t1.join()
        f.close()
        # it took at least 2 seconds to read
        # although nothing is there until .close()
        self.assertTrue(end - start >= 2)
        self.assertEqual(content, to_bytes(""))
        content = read_locked("test.txt")
        self.assertEqual(content, to_bytes("test ok"))

        f = LockedFile("test.txt", mode="wb")
        f.write(to_bytes("test ok"))
        t1 = threading.Thread(target=worker2, args=(f,))
        t1.start()
        start = int(time.time())
        content = read_locked("test.txt")
        end = int(time.time())
        t1.join()
        # it took at least 2 seconds to read
        # content is there because we called close()
        self.assertTrue(end - start >= 2)
        self.assertEqual(content, to_bytes("test ok"))

    @unittest.skipIf(IS_GAE, "GAE has no locks")
    def test_read_locked(self):
        def worker(fh):
            time.sleep(2)
            fh.close()

        f = LockedFile("test.txt", mode="wb")
        f.write(to_bytes("test ok"))
        t1 = threading.Thread(target=worker, args=(f,))
        t1.start()
        start = int(time.time())
        content = read_locked("test.txt")
        end = int(time.time())
        t1.join()
        # it took at least 2 seconds to read
        self.assertTrue(end - start >= 2)
        self.assertEqual(content, to_bytes("test ok"))

    @unittest.skipIf(IS_GAE, "GAE has no locks")
    def test_write_locked(self):
        def worker(fh):
            time.sleep(2)
            fh.close()

        f = open("test.txt", mode="wb")
        lock(f, LOCK_EX)
        t1 = threading.Thread(target=worker, args=(f,))
        t1.start()
        start = int(time.time())
        write_locked("test.txt", to_bytes("test ok"))
        end = int(time.time())
        t1.join()
        with open("test.txt") as g:
            content = g.read()
        # it took at least 2 seconds to read
        self.assertTrue(end - start >= 2)
        self.assertEqual(content, "test ok")

    def test_exception(self):
        self.assertRaises(RuntimeError, LockedFile, *["test.txt", "x"])

    def test_readline(self):
        f = LockedFile("test.txt", "wb")
        f.write(to_bytes("abc\n"))
        f.write(to_bytes("123\n"))
        f.close()
        f = LockedFile("test.txt", "rb")
        rl = f.readline()
        self.assertTrue(to_bytes("abc") in rl)
        rl = f.readline()
        self.assertTrue(to_bytes("123") in rl)
        f.close()
        f = LockedFile("test.txt", "rb")
        rls = f.readlines()
        f.close()
        self.assertEqual(len(rls), 2)



import re

from .._compat import PY2, integer_types, iteritems, long, to_unicode
from .._globals import IDENTITY
from ..utils import split_uri_args
from . import adapters, with_connection_or_raise
from .base import SQLAdapter


class Slicer(object):
    def rowslice(self, rows, minimum=0, maximum=None):
        if maximum is None:
            return rows[minimum:]
        return rows[minimum:maximum]


class MSSQL(SQLAdapter):
    dbengine = "mssql"
    drivers = ("pyodbc", "pytds")

    REGEX_DSN = "^.+$"
    REGEX_URI = (
        "^(?P<user>[^:@]+)(:(?P<password>[^@]*))?"
        r"@(?P<host>[^:/]+|\[[^\]]+\])(:(?P<port>\d+))?"
        "/(?P<db>[^?]+)"
        r"(\?(?P<uriargs>.*))?$"
    )

    def __init__(
        self,
        db,
        uri,
        pool_size=0,
        folder=None,
        db_codec="UTF-8",
        credential_decoder=IDENTITY,
        driver_args={},
        adapter_args={},
        srid=4326,
        after_connection=None,
    ):
        self.srid = srid
        super(MSSQL, self).__init__(
            db,
            uri,
            pool_size,
            folder,
            db_codec,
            credential_decoder,
            driver_args,
            adapter_args,
            after_connection,
        )

    def _initialize_(self):
        super(MSSQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        if "@" not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            self.dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL: %s" % self.uri)
            user = self.credential_decoder(m.group("user"))
            password = self.credential_decoder(m.group("password"))
            if password is None:
                password = ""
            host = m.group("host")
            db = m.group("db")
            port = m.group("port") or "1433"
            # Parse the optional uri name-value arg pairs after the '?'
            # (in the form of arg1=value1&arg2=value2&...)
            # (drivers like FreeTDS insist on uppercase parameter keys)
            argsdict = {"DRIVER": "{SQL Server}"}
            uriargs = m.group("uriargs")
            if uriargs:
                for argkey, argvalue in split_uri_args(
                    uriargs, separators="&", need_equal=True
                ).items():
                    argsdict[argkey.upper()] = argvalue
            uriargs = ";".join(["%s=%s" % (ak, av) for (ak, av) in iteritems(argsdict)])
            self.dsn = "SERVER=%s;PORT=%s;DATABASE=%s;UID=%s;PWD=%s;%s" % (
                host,
                port,
                db,
                user,
                password,
                uriargs,
            )

    def connector(self):
        return self.driver.connect(self.dsn, **self.driver_args)

    def lastrowid(self, table):
        self.execute("SELECT SCOPE_IDENTITY();")
        return int(self.cursor.fetchone()[0])


@adapters.register_for("mssql")
class MSSQL1(MSSQL, Slicer):
    pass


@adapters.register_for("mssql3")
class MSSQL3(MSSQL):
    pass


@adapters.register_for("mssql4")
class MSSQL4(MSSQL):
    pass


class MSSQLN(MSSQL):
    def represent(self, obj, field_type):
        rv = super(MSSQLN, self).represent(obj, field_type)
        if field_type in ("string", "text", "json") and rv.startswith("'"):
            rv = "N" + rv
        return rv

    @with_connection_or_raise
    def execute(self, *args, **kwargs):
        if PY2:
            args = list(args)
            args[0] = to_unicode(args[0])
        return super(MSSQLN, self).execute(*args, **kwargs)


@adapters.register_for("mssqln", "mssql2")
class MSSQL1N(MSSQLN, Slicer):
    pass


@adapters.register_for("mssql3n")
class MSSQL3N(MSSQLN):
    pass


@adapters.register_for("mssql4n")
class MSSQL4N(MSSQLN):
    pass


@adapters.register_for("pytds")
class PyTDS(MSSQL):
    def _initialize_(self):
        super(MSSQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        if "@" not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            self.dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL: %s" % self.uri)
            self.dsn = m.group("host")
            self.driver_args.update(
                user=self.credential_decoder(m.group("user")),
                password=self.credential_decoder(m.group("password")) or "",
                database=m.group("db"),
                port=m.group("port") or "1433",
            )

    def connector(self):
        return self.driver.connect(self.dsn, **self.driver_args)


@adapters.register_for("vertica")
class Vertica(MSSQL1):
    def lastrowid(self, table):
        self.execute("SELECT SCOPE_IDENTITY();")
        return int(self.cursor.fetchone()[0])


@adapters.register_for("sybase")
class Sybase(MSSQL1):
    dbengine = "sybase"

    def _initialize_(self):
        super(MSSQL, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        if "@" not in ruri:
            m = re.match(self.REGEX_DSN, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL")
            dsn = m.group()
        else:
            m = re.match(self.REGEX_URI, ruri)
            if not m:
                raise SyntaxError("Invalid URI string in DAL: %s" % self.uri)
            user = self.credential_decoder(m.group("user"))
            password = self.credential_decoder(m.group("password"))
            if password is None:
                password = ""
            host = m.group("host")
            db = m.group("db")
            port = m.group("port") or "1433"
            self.dsn = "sybase:host=%s:%s;dbname=%s" % (host, port, db)
            self.driver_args.update(
                user=self.credential_decoder(user),
                passwd=self.credential_decoder(password),
            )



from datetime import datetime

from .._compat import PY2, integer_types
from ..adapters.mongo import Mongo, MongoBlob
from ..helpers.classes import Reference
from . import Parser, before_parse, for_type, parsers

long = integer_types[-1]


@parsers.register_for(Mongo)
class MongoParser(Parser):
    @for_type("id")
    def _id(self, value):
        if isinstance(value, self.adapter.ObjectId):
            return int(str(value), 16)
        return int(value)

    @for_type("blob")
    def _blob(self, value):
        return MongoBlob.decode(value) if PY2 else value

    @before_parse("reference")
    def reference_extras(self, field_type):
        return {"referee": field_type[10:].strip()}

    @for_type("reference")
    def _reference(self, value, referee):
        if isinstance(value, self.adapter.ObjectId):
            value = int(str(value), 16)
        if "." not in referee:
            value = Reference(value)
            value._table, value._record = self.adapter.db[referee], None
        return value

    @before_parse("list:reference")
    def referencelist_extras(self, field_type):
        return {"field_type": field_type}

    @for_type("list:reference")
    def _list_references(self, value, field_type):
        return [self.registered["reference"](el, field_type[5:]) for el in value]

    @for_type("date")
    def _date(self, value):
        if isinstance(value, datetime):
            return value.date()
        return value

    @for_type("time")
    def _time(self, value):
        if isinstance(value, datetime):
            return value.time()
        return value



# -*- coding: utf-8 -*-
# pylint: disable=no-member

import os
import threading

from ._compat import itervalues
from ._globals import GLOBAL_LOCKER, THREAD_LOCAL
from ._load import OrderedDict


class ConnectionPool(object):
    POOLS = {}
    check_active_connection = True

    def __init__(self):
        self._first_connection = False

    @property
    def _connection_uname_(self):
        return "_pydal_connection_%s_%s" % (id(self), os.getpid())

    @property
    def _cursors_uname_(self):
        return "_pydal_cursor_%s_%s" % (id(self), os.getpid())

    @staticmethod
    def set_folder(folder):
        THREAD_LOCAL._pydal_folder_ = folder

    @property
    def connection(self):
        return self.get_connection()

    def get_connection(self, use_pool=True):
        """
        if `self.pool_size>0` it will try pull the connection from the pool
        if the connection is not active (closed by db server) it will loop
        if not `self.pool_size` or no active connections in pool makes a new one
        """
        # check we we have a connection for this process/thread/object id
        connection = getattr(THREAD_LOCAL, self._connection_uname_, None)
        # if so, return it
        if connection is not None:
            return connection

        # if not and we have a pool
        if use_pool and self.pool_size:
            try:
                GLOBAL_LOCKER.acquire()
                pool = ConnectionPool.POOLS.get(self.uri, [])
                ConnectionPool.POOLS[self.uri] = pool
                # pop from the pool until we find a valid connection
                while connection is None and pool:
                    connection = pool.pop()
                    try:
                        self.set_connection(connection, run_hooks=False)
                    except:
                        connection = None
            finally:
                GLOBAL_LOCKER.release()

        # if still no connection, make a new one and run the hooks
        # note we serialize actual connections to protect hooks
        if connection is None:
            connection = self.connector()
            self.set_connection(connection, run_hooks=True)

        return connection

    def set_connection(self, connection, run_hooks=False):
        # store the connection in the thread local object and run hooks (optional)
        setattr(THREAD_LOCAL, self._connection_uname_, connection)
        if connection:
            setattr(THREAD_LOCAL, self._cursors_uname_, connection.cursor())
            # run hooks
            if run_hooks:
                self.after_connection_hook()
            # some times we want to check the connection is still good
            if self.check_active_connection:
                self.test_connection()
        else:
            setattr(THREAD_LOCAL, self._cursors_uname_, None)

    def reset_cursor(self):
        """get a new cursor for the existing connection"""
        setattr(THREAD_LOCAL, self._cursors_uname_, self.connection.cursor())

    @property
    def cursor(self):
        """retrieve the cursor of the connection"""
        return getattr(THREAD_LOCAL, self._cursors_uname_)

    def _clean_tlocals(self):
        """delete cusor and connection from the thead local"""
        delattr(THREAD_LOCAL, self._cursors_uname_)
        delattr(THREAD_LOCAL, self._connection_uname_)

    def close(self, action="commit", really=True):
        """if we have an action (commit, rollback), try to execute it"""
        # if the connection was never established, nothing to do
        if getattr(THREAD_LOCAL, self._connection_uname_, None) is None:
            return
        # try commit or rollback
        succeeded = True
        if action:
            try:
                if callable(action):
                    action(self)
                else:
                    getattr(self, action)()
            except:
                #: connection had some problems, we want to drop it
                succeeded = False
        # close the cursor
        self.cursor.close()
        # if we have pools, we should recycle the connection (but only when
        # we succeded in `action`, if any and `len(pool)` is good)
        if self.pool_size and succeeded:
            try:
                GLOBAL_LOCKER.acquire()
                pool = ConnectionPool.POOLS[self.uri]
                if len(pool) < int(self.pool_size):
                    pool.append(self.connection)
                    really = False
            finally:
                GLOBAL_LOCKER.release()
        #: closing the connection when we `really` want to, in particular:
        #    - when we had an exception running `action`
        #    - when we don't have pools
        #    - when we have pools but they're full
        if really:
            try:
                self.close_connection()
            except:
                pass
        #: always unset `connection` attribute
        self.set_connection(None)

    @staticmethod
    def close_all_instances(action):
        """to close cleanly databases in a multithreaded environment"""
        dbs = getattr(THREAD_LOCAL, "_pydal_db_instances_", {}).items()
        for db_uid, db_group in dbs:
            for db in db_group:
                if hasattr(db, "_adapter"):
                    db._adapter.close(action)
        getattr(THREAD_LOCAL, "_pydal_db_instances_", {}).clear()
        getattr(THREAD_LOCAL, "_pydal_db_instances_zombie_", {}).clear()
        if callable(action):
            action(None)

    def _find_work_folder(self):
        self.folder = getattr(THREAD_LOCAL, "_pydal_folder_", "")

    def after_connection_hook(self):
        """Hook for the after_connection parameter"""
        # some work must be done on first connection only
        if not self._first_connection:
            self._after_first_connection()
            self._first_connection = True
        # handle user specified hooks if present
        if callable(self._after_connection):
            self._after_connection(self)
        # handle global adapter hooks
        self.after_connection()

    def after_connection(self):
        # this it is supposed to be overloaded by adapters
        pass

    def _after_first_connection(self):
        """called only after first connection"""
        pass

    def reconnect(self):
        """legacy method - no longer needed"""
        self.close()
        self.get_connection()



# -*- coding: utf-8 -*-
import copy
import marshal
import struct
import threading
import time
import traceback

from .._compat import (
    PY2,
    copyreg,
    exists,
    implements_bool,
    iteritems,
    iterkeys,
    itervalues,
    long,
    to_bytes,
)
from .._globals import THREAD_LOCAL
from .serializers import serializers


class cachedprop(object):
    #: a read-only @property that is only evaluated once.
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


@implements_bool
class BasicStorage(object):
    def __init__(self, *args, **kwargs):
        return self.__dict__.__init__(*args, **kwargs)

    def __getitem__(self, key):
        return self.__dict__.__getitem__(str(key))

    __setitem__ = object.__setattr__

    def __delitem__(self, key):
        try:
            delattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __bool__(self):
        return len(self.__dict__) > 0

    __iter__ = lambda self: self.__dict__.__iter__()

    __str__ = lambda self: self.__dict__.__str__()

    __repr__ = lambda self: self.__dict__.__repr__()

    has_key = __contains__ = lambda self, key: key in self.__dict__

    def get(self, key, default=None):
        return self.__dict__.get(key, default)

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def iterkeys(self):
        return iterkeys(self.__dict__)

    def values(self):
        return self.__dict__.values()

    def itervalues(self):
        return itervalues(self.__dict__)

    def items(self):
        return self.__dict__.items()

    def iteritems(self):
        return iteritems(self.__dict__)

    pop = lambda self, *args, **kwargs: self.__dict__.pop(*args, **kwargs)

    clear = lambda self, *args, **kwargs: self.__dict__.clear(*args, **kwargs)

    copy = lambda self, *args, **kwargs: self.__dict__.copy(*args, **kwargs)


def pickle_basicstorage(s):
    return BasicStorage, (dict(s),)


copyreg.pickle(BasicStorage, pickle_basicstorage)


class OpRow(object):
    __slots__ = ("_table", "_fields", "_values")

    def __init__(self, table):
        object.__setattr__(self, "_table", table)
        object.__setattr__(self, "_fields", {})
        object.__setattr__(self, "_values", {})

    def set_value(self, key, value, field=None):
        self._values[key] = value
        self._fields[key] = self._fields.get(key, field or self._table[key])

    def del_value(self, key):
        del self._values[key]
        del self._fields[key]

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        return self.set_value(key, value)

    def __delitem__(self, key):
        return self.del_value(key)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError

    def __setattr__(self, key, value):
        return self.set_value(key, value)

    def __delattr__(self, key):
        return self.del_value(key)

    def __iter__(self):
        return self._values.__iter__()

    def __contains__(self, key):
        return key in self._values

    def get(self, key, default=None):
        try:
            rv = self[key]
        except KeyError:
            rv = default
        return rv

    def keys(self):
        return self._values.keys()

    def iterkeys(self):
        return iterkeys(self._values)

    def values(self):
        return self._values.values()

    def itervalues(self):
        return itervalues(self._values)

    def items(self):
        return self._values.items()

    def iteritems(self):
        return iteritems(self._values)

    def op_values(self):
        return [(self._fields[key], value) for key, value in iteritems(self._values)]

    def __repr__(self):
        return "<OpRow %s>" % repr(self._values)


class Serializable(object):
    def as_dict(self, flat=False, sanitize=True):
        return self.__dict__

    def as_xml(self, sanitize=True):
        return serializers.xml(self.as_dict(flat=True, sanitize=sanitize))

    def as_json(self, sanitize=True):
        return serializers.json(self.as_dict(flat=True, sanitize=sanitize))

    def as_yaml(self, sanitize=True):
        return serializers.yaml(self.as_dict(flat=True, sanitize=sanitize))


class Reference(int):
    def __allocate(self):
        if not self._record:
            self._record = self._table[int(self)]
        if not self._record:
            raise RuntimeError(
                "Using a recursive select but encountered a broken "
                + "reference: %s %d" % (self._table, int(self))
            )

    def __getattr__(self, key, default=None):
        if key == "id":
            return int(self)
        if key in self._table:
            self.__allocate()
        if self._record:
            # to deal with case self.update_record()
            return self._record.get(key, default)
        else:
            return None

    def get(self, key, default=None):
        return self.__getattr__(key, default)

    def __setattr__(self, key, value):
        if key.startswith("_"):
            long.__setattr__(self, key, value)
            return
        self.__allocate()
        self._record[key] = value

    def __getitem__(self, key):
        if key == "id":
            return int(self)
        self.__allocate()
        return self._record.get(key, None)

    def __setitem__(self, key, value):
        self.__allocate()
        self._record[key] = value


def Reference_unpickler(data):
    return marshal.loads(data)


def Reference_pickler(data):
    try:
        marshal_dump = marshal.dumps(int(data))
    except AttributeError:
        marshal_dump = "i%s" % struct.pack("<i", int(data))
    return (Reference_unpickler, (marshal_dump,))


copyreg.pickle(Reference, Reference_pickler, Reference_unpickler)


class SQLCallableList(list):
    def __call__(self):
        return copy.copy(self)


class SQLALL(object):
    """
    Helper class providing a comma-separated string having all the field names
    (prefixed by table name and '.')

    normally only called from within gluon.dal
    """

    def __init__(self, table):
        self._table = table

    def __str__(self):
        return ", ".join([str(field) for field in self._table])


class SQLCustomType(object):
    """
    Allows defining of custom SQL types

    Args:
        type: the web2py type (default = 'string')
        native: the backend type
        encoder: how to encode the value to store it in the backend
        decoder: how to decode the value retrieved from the backend
        validator: what validators to use ( default = None, will use the
            default validator for type)

    Example::
        Define as:

            decimal = SQLCustomType(
                type ='double',
                native ='integer',
                encoder =(lambda x: int(float(x) * 100)),
                decoder = (lambda x: Decimal("0.00") + Decimal(str(float(x)/100)) )
                )

            db.define_table(
                'example',
                Field('value', type=decimal)
                )

    """

    def __init__(
        self,
        type="string",
        native=None,
        encoder=None,
        decoder=None,
        validator=None,
        _class=None,
        widget=None,
        represent=None,
    ):
        self.type = type
        self.native = native
        self.encoder = encoder or (lambda x: x)
        self.decoder = decoder or (lambda x: x)
        self.validator = validator
        self._class = _class or type
        self.widget = widget
        self.represent = represent

    def startswith(self, text=None):
        try:
            return self.type.startswith(self, text)
        except TypeError:
            return False

    def endswith(self, text=None):
        try:
            return self.type.endswith(self, text)
        except TypeError:
            return False

    def __getslice__(self, a=0, b=100):
        return None

    def __getitem__(self, i):
        return None

    def __str__(self):
        return self._class


class RecordOperator(object):
    def __init__(self, colset, table, id):
        self.colset, self.db, self.tablename, self.id = (
            colset,
            table._db,
            table._tablename,
            id,
        )

    def __call__(self):
        pass


class RecordUpdater(RecordOperator):
    def __call__(self, **fields):
        colset, db, tablename, id = self.colset, self.db, self.tablename, self.id
        table = db[tablename]
        newfields = fields or dict(colset)
        for fieldname in list(newfields.keys()):
            if fieldname not in table.fields or table[fieldname].type == "id":
                del newfields[fieldname]
        table._db(table._id == id, ignore_common_filters=True).update(**newfields)
        colset.update(newfields)
        return colset


class RecordDeleter(RecordOperator):
    def __call__(self):
        return self.db(self.db[self.tablename]._id == self.id).delete()


class MethodAdder(object):
    def __init__(self, table):
        self.table = table

    def __call__(self):
        return self.register()

    def __getattr__(self, method_name):
        return self.register(method_name)

    def register(self, method_name=None):
        def _decorated(f):
            instance = self.table
            import types

            if PY2:
                method = types.MethodType(f, instance, instance.__class__)
            else:
                method = types.MethodType(f, instance)
            name = method_name or f.func_name
            setattr(instance, name, method)
            return f

        return _decorated


class FakeCursor(object):
    """
    The Python Database API Specification has a cursor() method, which
    NoSql drivers generally don't support.  If the exception in this
    function is taken then it likely means that some piece of
    functionality has not yet been implemented in the driver. And
    something is using the cursor.

    https://www.python.org/dev/peps/pep-0249/
    """

    def warn_bad_usage(self, attr):
        raise Exception("FakeCursor.%s is not implemented" % attr)

    def __getattr__(self, attr):
        self.warn_bad_usage(attr)

    def __setattr__(self, attr, value):
        self.warn_bad_usage(attr)

    def close(self):
        return


class NullCursor(FakeCursor):
    lastrowid = 1

    def __getattr__(self, attr):
        return lambda *a, **b: []


class FakeDriver(BasicStorage):
    def __init__(self, *args, **kwargs):
        super(FakeDriver, self).__init__(*args, **kwargs)
        self._build_cursor_()

    def _build_cursor_(self):
        self._fake_cursor_ = FakeCursor()

    def cursor(self):
        return self._fake_cursor_

    def close(self):
        return None

    def commit(self):
        return None

    def __str__(self):
        state = ["%s=%r" % (attribute, value) for (attribute, value) in self.items()]
        return "\n".join(state)


class NullDriver(FakeDriver):
    def _build_cursor_(self):
        self._fake_cursor_ = NullCursor()


class ExecutionHandler(object):
    def __init__(self, adapter):
        self.adapter = adapter

    def before_execute(self, command):
        pass

    def after_execute(self, command):
        pass


class TimingHandler(ExecutionHandler):
    MAXSTORAGE = 100

    def _timings(self):
        THREAD_LOCAL._pydal_timings_ = getattr(THREAD_LOCAL, "_pydal_timings_", [])
        return THREAD_LOCAL._pydal_timings_

    @property
    def timings(self):
        return self._timings()

    def before_execute(self, command):
        self.t = time.time()

    def after_execute(self, command):
        dt = time.time() - self.t
        self.timings.append((command, dt))
        del self.timings[: -self.MAXSTORAGE]


class DatabaseStoredFile:
    web2py_filesystems = set()

    def escape(self, obj):
        return self.db._adapter.escape(obj)

    @staticmethod
    def try_create_web2py_filesystem(db):
        if db._uri not in DatabaseStoredFile.web2py_filesystems:
            if db._adapter.dbengine not in ("mysql", "postgres", "sqlite"):
                raise NotImplementedError(
                    "DatabaseStoredFile only supported by mysql, potresql, sqlite"
                )
            blobType = "BYTEA" if db._adapter.dbengine == "postgres" else "BLOB"
            sql = (
                "CREATE TABLE IF NOT EXISTS web2py_filesystem (path VARCHAR(255), content %(blobType)s, PRIMARY KEY(path));"
                % {"blobType": blobType}
            )
            if db._adapter.dbengine == "mysql":
                sql = sql[:-1] + " ENGINE=InnoDB;"
            db.executesql(sql)
            DatabaseStoredFile.web2py_filesystems.add(db._uri)

    def __init__(self, db, filename, mode):
        if db._adapter.dbengine not in ("mysql", "postgres", "sqlite"):
            raise RuntimeError(
                "only MySQL/Postgres/SQLite can store metadata .table files"
                + " in database for now"
            )
        self.db = db
        self.filename = filename
        self.mode = mode
        DatabaseStoredFile.try_create_web2py_filesystem(db)
        self.p = 0
        self.data = b""
        if mode in ("r", "rw", "rb", "a", "ab"):
            query = "SELECT content FROM web2py_filesystem WHERE path='%s'" % filename
            rows = self.db.executesql(query)
            if rows:
                self.data = to_bytes(rows[0][0])
            elif exists(filename):
                datafile = open(filename, "rb")
                try:
                    self.data = datafile.read()
                finally:
                    datafile.close()
            elif mode in ("r", "rw", "rb"):
                raise RuntimeError("File %s does not exist" % filename)

    def read(self, bytes=None):
        if bytes is None:
            bytes = len(self.data)
        data = self.data[self.p : self.p + bytes]
        self.p += len(data)
        return data

    def readinto(self, bytes):
        return self.read(bytes)

    def readline(self):
        i = self.data.find(b"\n", self.p) + 1
        if i > 0:
            data, self.p = self.data[self.p : i], i
        else:
            data, self.p = self.data[self.p :], len(self.data)
        return data

    def write(self, data):
        self.data += data

    def close_connection(self):
        if self.db is not None:
            self.db.executesql(
                "DELETE FROM web2py_filesystem WHERE path='%s'" % self.filename
            )
            placeholder = "?" if self.db._adapter.dbengine == "sqlite" else "%s"
            query = (
                "INSERT INTO web2py_filesystem(path,content) VALUES (%(placeholder)s, %(placeholder)s)"
                % {"placeholder": placeholder}
            )
            args = (self.filename, self.data)
            self.db.executesql(query, args)
            self.db.commit()
            self.db = None

    def close(self):
        self.close_connection()

    @staticmethod
    def is_operational_error(db, error):
        if not hasattr(db._adapter.driver, "OperationalError"):
            return None
        return isinstance(error, db._adapter.driver.OperationalError)

    @staticmethod
    def is_programming_error(db, error):
        if not hasattr(db._adapter.driver, "ProgrammingError"):
            return None
        return isinstance(error, db._adapter.driver.ProgrammingError)

    @staticmethod
    def exists(db, filename):
        if exists(filename):
            return True

        DatabaseStoredFile.try_create_web2py_filesystem(db)

        query = "SELECT path FROM web2py_filesystem WHERE path='%s'" % filename
        try:
            if db.executesql(query):
                return True
        except Exception as e:
            if not (
                DatabaseStoredFile.is_operational_error(db, e)
                or DatabaseStoredFile.is_programming_error(db, e)
            ):
                raise
            # no web2py_filesystem found?
            tb = traceback.format_exc()
            db.logger.error("Could not retrieve %s\n%s" % (filename, tb))
        return False
